{
  This unit handles graphicsmodes 10h,12h,13h and unchained 13h. All these
  procedures are written in 100% inline assembler by Vulture/Outlaw Triad,
  except for the hires getpixel function which, partially, came from those
  cool SWAG files.

  If something nice is written using this code, please mail me at:
  " comma400@tem.nhl.nl " as I'm very interested!

      -Vulture/Outlaw Triad-
}

Unit Gfx;                       { Vga graphics unit; 1996/1997/1998 }

INTERFACE

Const VGA = $a000;              { Vga segment }
      CrtC_Index = $03d4;       { Crt controller index }
      Gfx_Index = $03ce;        { Graphics controller index }
      Seq_Index = $03c4;        { Sequencer adress index }
      ScrWidth: Byte = 80;      { Screenwidth in bytes (hires modes) }

Var   RGB_Values: Array[0..767] of Byte;      { Array for all RGB values }

Procedure VideoMode(Mode: Byte);
Procedure SetPixel(X,Y:Integer;Color:Byte;Where:Word);
Function  GetPixel(X,Y:Integer; Where: Word): Byte;
Procedure ClearScreen(Color:Byte;Where:Word);
Procedure Copy_Block(x1,y1,x2,y2,dx1,dy1,WhereFrom,WhereTo:Word);
Procedure H_Line(x1,x2,y:Word;Color:Byte;Where:Word);
Procedure FlipPage(Source,Dest:Word);
Procedure WaitRetrace;
Procedure StartRetrace;
Procedure EndRetrace;
Procedure Set_Unchained_Mode(Size: Byte);
Procedure Clear_Unchained;
Procedure SetPixel_C4(X,Y:Integer; Size,Color:Byte);
Function  GetPixel_C4(X,Y:Integer; Size:Byte): Byte;
Procedure SetPixel_Hires(X,Y: Word; Color: Byte);
Function  GetPixel_Hires(X,Y: Word): Byte;
Procedure Clear_Hires(Color: Byte);
Procedure Set_Hires_Color(Color: Byte);
Procedure Set_Lineair_Palette;
Procedure Move_To(X,Y: Word);
Procedure Set_Line(LineValue: Word);
Procedure Copy_Vga(Source,Dest,Count: Word);
Procedure Set_Plane(Plane: Byte);
Procedure Set_Height(Height: Byte);
Procedure Set_WriteMode(Mode: Byte);
Procedure Set_Color(Color,R,G,B: Byte);
Procedure Get_Color(Color: Byte; Var R,G,B: Byte);
Procedure Set_PaletteReg(Reg,Val: Byte);

IMPLEMENTATION

Procedure VideoMode(Mode: Byte); Assembler;
Asm
    xor     ah,ah
    mov     al,Mode             { Load vgamode }
    int     10h
End;

Procedure SetPixel(X,Y:Integer;Color:Byte;Where:Word); Assembler;
Asm
    mov     ax,Where
    mov     es,ax
    mov     di,Y
    mov     ax,Y
    shl     di,8
    shl     ax,6
    add     di,ax
    add     di,X
    mov     al,Color
    mov     byte ptr es:[di],al
End;

Function GetPixel(X,Y:Integer; Where: Word): Byte; Assembler;
Asm
    mov     ax,Where
    mov     es,ax
    mov     di,Y
    mov     ax,Y
    shl     di,8
    shl     ax,6
    add     di,ax
    add     di,X
    xor     ax,ax
    mov     al,byte ptr es:[di]
End;

Procedure ClearScreen(Color:Byte;Where:Word); Assembler;
Asm
    mov     ax,Where
    mov     es,ax          { ES points to VGA or vitual screen }
    xor     di,di          { Start at begin of screen }
    cld                    { Clear direction flag (di=increased) }
    mov     al,Color       { Set color in AL }
    mov     ah,al
    mov     cx,32000       { Do the entire screen }
    rep     stosw          { Store the word in ax to es:[di] }
End;

{ This routine copies blocks with a width which is dividable by 4! }
Procedure Copy_Block(x1,y1,x2,y2,dx1,dy1,WhereFrom,WhereTo:Word); Assembler;
Var Subs: Word;
Asm
    push    ds
    mov     ax,WhereFrom
    mov     ds,ax           { ds = source segment }
    mov     ax,y1
    mov     bx,ax
    shl     ax,8
    shl     bx,6
    add     ax,bx
    add     ax,x1
    mov     si,ax           { ds:si = final source }

    mov     ax,WhereTo
    mov     es,ax           { es = destination segment }
    mov     ax,dy1
    mov     bx,ax
    shl     ax,8
    shl     bx,6
    add     ax,bx
    add     ax,dx1
    mov     di,ax           { es:di = final destination }

    mov     cx,x2
    sub     cx,x1           { cx = # of bytes per scanline }
    mov     ax,320
    sub     ax,cx
    mov     Subs,ax         { ax = # of bytes to add after copying }
    shr     cx,2            { cx = cx / 4 (for dword copy }
    mov     dx,cx           { Save it }
    mov     bx,y2
    sub     bx,y1           { bx = # of scanlines to process }

@Y_Loop:
    mov     cx,dx           { Set # of dwords to copy }
    db      66h
    rep     movsw           { Copy using dwords }
    add     si,Subs         { Next offset in source }
    add     di,Subs         { Next offset in destination }
    dec     bx
    jnz     @Y_Loop         { Do all scanlines }

    pop     ds
End;

Procedure H_Line(x1,x2,y:Word;Color:Byte;Where:Word); Assembler;
Asm
    mov   bx,x1
    mov   cx,x2
    cmp   bx,cx
    jb    @Line_Ok
    je    @Quit_Line
    xchg  bx,cx                 { bx=x1   cx=x2 }
@Line_Ok:

    mov   ax,Where
    mov   es,ax
    mov   ax,y
    mov   di,ax
    shl   ax,8
    shl   di,6
    add   di,ax                 { di=y*320 }
    add   di,bx                 { di=y*320+x }

    mov   al,Color
    mov   ah,al
    cld
    sub   cx,bx                 { cx=length of line }
    shr   cx,1
    jnc   @Start_Fill
    stosb                       { Plot extra pixel (odd # pixels) }
@Start_Fill:
    rep   stosw                 { Plot all remaining pixels (even # pixels) }
@Quit_Line:
End;

Procedure H_Line_Hi(x1,x2,y:Word;Color:Byte;Where:Word); Assembler;
Asm
    mov     dx,3ceh             { Graphic Controller address register port }
    mov     ax,0f01h            { Index 01h  (enable set/reset register) }
    out     dx,ax
    xor     al,al               { Index 00h  (set/reset register) }
    mov     ah,Color            { Color (0 to 15) }
    out     dx,ax

    mov     cx,x1               { Fill first byte }
    and     cx,00000111b        { x and 7 }
    mov     ah,10000000b        { Setup ah for bitselection }
    shr     ah,cl               { ah= select bit (pixel) }
    mov     al,08h              { Select Bit Mask Register }
    out     dx,ax

    mov     cx,x2               { Fill all bytes }
    sub     cx,x1
    shr     cx,3                { x div 8 (# of complete bytes to fill) }
    mov     ah,11111111b
    mov     al,08h
    out     dx,ax
@Hori:
    mov     byte ptr es:[di],ah
    inc     di
    loop    @Hori

    mov     cx,x2               { Fill last byte }
    xor     ch,ch
    and     cx,00000111b        { x and 7 }
    mov     ah,10000000b        { Setup ah for bitselection }
    shr     ah,cl               { ah= select bit (pixel) }
@Set2:
    mov     bl,11000000b
    shr     bl,1
    or      ah,bl
    loop    @Set2

    mov     al,08h              { Select Bit Mask Register }
    out     dx,ax

End;

Procedure FlipPage(Source,Dest:Word); Assembler;
Asm
    push    ds
    mov     ax,[Source]
    mov     ds,ax        { ds points to source segment }
    xor     si,si        { Start at begin of source }
    mov     ax,[Dest]
    mov     es,ax        { es points to destination segment }
    xor     di,di        { Start at begin of destination }
    mov     cx,16000     { # of dwords }
    db      $66
    rep     movsw        { Copy ds:si to es:di using dwords }
    pop     ds
End;

Procedure WaitRetrace; Assembler;
Asm
    mov     dx,3dah
@Vrt:
    in      al,dx
    test    al,1000b
    jnz     @Vrt
@NoVrt:
    in      al,dx
    test    al,1000b
    jz      @NoVrt
End;

Procedure StartRetrace; Assembler;
Asm
    mov     dx,3dah
@Vrt:
    in      al,dx
    test    al,1000b
    jnz     @Vrt
End;

Procedure EndRetrace; Assembler;
Asm
    mov     dx,3dah
@NoVrt:
    in      al,dx
    test    al,1000b
    jz      @NoVrt
End;

Procedure Set_Unchained_Mode(Size: Byte); Assembler; { Set unchained mode 13h }
Asm
    mov   ax,0013h
    int   10h                   { Set mode 13h (320*200*256) }

    mov   dx,03c4h              { Sequencer Address Register }
    mov   ax,0604h              { 06h = 00000110b }
    out   dx,ax                 { Disable chain4 bit }

    mov   dx,03d4h              { CrtC register }
    mov   ax,0014h
    out   dx,ax                 { Turn off double word mode }
    mov   ax,0e317h
    out   dx,ax                 { Turn on byte mode }

    mov   dx,03c4h              { Sequencer Address Register }
    mov   ax,00f02h             { Map Mask Register (select all planes) }
    out   dx,ax
    mov   ax,0a000h             { VGA segment }
    mov   es,ax
    xor   di,di                 { es:di => screen memory }
    xor   ax,ax
    mov   cx,32768              { 32768 words * 2 = 65536 bytes = 1 plane }
    rep   stosw                 { Clear all 4 planes using wordwrites }

    mov   dx,3d4h
    mov   al,13h                { Index Logical Screen Width }
    mov   ah,[Size]             { Load size (40,80,160) }
    out   dx,ax
End;

Procedure Clear_Unchained; Assembler;   { Clears unchained screen (black) }
Asm
    mov   dx,Seq_Index          { Select port 03c4h }
    mov   ax,00f02h             { Map Mask Register (plane select) }
    out   dx,ax                 { Select all planes }
    mov   ax,0a000h             { VGA segment }
    mov   es,ax
    xor   di,di                 { es:di => vga memory }
    xor   ax,ax
    mov   cx,32768              { 32768 (words) * 2 = 65536 bytes = 1 plane }
    cld
    rep   stosw                 { Clear all 4 planes using wordwrites }
End;

Procedure SetPixel_C4(X,Y:Integer; Size,Color:Byte); Assembler;
Asm
   mov    ax,[Y]
   xor    bx,bx
   mov    bl,[Size]
   imul   bx                       { Now ax:=ax*bx => Y:=Y*Size }
   shl    ax,1                     { Multiply by 2 => Y:=Y*2 (Dunno why) }
   mov    bx,ax                    { Save Y-value in bx }
   mov    ax,[X]
   mov    cx,ax
   shr    ax,2                     { Divide ax by 4 }
   add    bx,ax                    { Y:=Y+X   bx contains final location }
   and    cx,00000011b             { cl:=0,1,2 or 3  (x mod 4) }
   mov    ah,00000001b             { Start value (plane 0) }
   shl    ah,cl                    { Eg plane 2: 00000001 => 00000100 }
   mov    dx,3c4h                  { Sequencer Register }
   mov    al,2                     { Select index 2 => plane enable }
   out    dx,ax                    { Enable selected plane }
   mov    ax,0a000h
   mov    es,ax                    { es points to VGA }
   mov    al,[Color]               { Color in al }
   mov    es:[bx],al               { Place it }
End;

Function GetPixel_C4(X,Y:Integer;Size:Byte): Byte; Assembler;
Asm
   mov    ax,[Y]                   { Put Y in ax }
   xor    bx,bx                    { Reset bx to zero }
   mov    bl,[Size]                { Size in bl (it's a Byte!) }
   imul   bx                       { Now ax:=ax*bx => Y:=Y*Size }
   shl    ax,1                     { Multiply by 2 => Y:=Y*2 (Dunno why) }
   mov    bx,ax                    { Save Y-value in bx }
   mov    ax,[X]                   { Move X-value into ax }
   mov    cx,ax                    { And in cx }
   shr    ax,2                     { Divide ax by 4  (x div 4) }
   add    bx,ax                    { Y:=Y+X   bx contains final location }
   and    cl,00000011b             { cl:=0,1,2 or 3  (x mod 4) }
   mov    dx,03ceh                 { Gfx controller }
   mov    ah,cl
   mov    al,4                     { Select index 2 => readplane enable }
   out    dx,ax                    { Enable selected plane }
   mov    ax,0a000h                { VGA segment }
   mov    es,ax                    { es points to VGA }
   xor    ax,ax
   mov    al,byte ptr es:[bx]      { Read it }
End;

Procedure SetPixel_Hires(X,Y: Word; Color: Byte); Assembler;
Asm
{ === Set color of pixel === }
    mov     dx,3ceh             { Graphic Controller address register port }
    mov     ax,0f01h            { Index 01h  (enable set/reset register) }
    out     dx,ax
    xor     al,al               { Index 00h  (set/reset register) }
    mov     ah,Color            { Color (0 to 15) }
    out     dx,ax

{ === Mask out incorrect bits === }
    mov     cx,X
    and     cx,00000111b        { x and 7 }
    mov     ah,10000000b        { Setup ah for bitselection }
    shr     ah,cl               { ah= select bit (pixel) }
    mov     al,08h              { Select Bit Mask Register }
    out     dx,ax

{ === Y*80 === }                { Can be altered for other screenwidths }
    mov     bx,Y                { bx=Y }
    mov     cx,bx               { cx=Y }
    shl     bx,6                { bx=bx*64 }
    shl     cx,4                { cx=cx*16 }
    add     bx,cx               { Y=Y*80  (80*8=640 pixels on 1 line) }

{ === Finalize position === }
    mov     ax,X                { ax=X }
    shr     ax,3                { X=X/8 (which byte to write to on scanline) }
    add     bx,ax               { Y=Y+X => bx=final position }

{ === Place pixel === }
    mov     ax,0a000h
    mov     es,ax
    mov     al,es:[bx]          { Dummy read from VGA (read latches) }
    mov     es:[bx],al          { Place dot }
End;

Function GetPixel_Hires(X,Y: Word): Byte; Assembler;
Asm
    mov     ax,y
    mov     bx,ax
    shl     ax,6
    shl     bx,4
    add     ax,bx               { ax=y*80 }

    mov     si,x
    mov     cx,si
    shr     si,3
    add     si,ax
    and     cl,$07
    xor     cl,$07
    mov     ch,1
    shl     ch,cl
    mov     ax,$A000
    mov     es,ax
    mov     dx,$03CE
    mov     ax,772
    xor     bl,bl
@Find:
    out     dx,ax
    mov     bh,es:[si]
    and     bh,ch
    neg     bh
    rol     bx,$0001
    dec     ah
    jge     @Find
    mov     al,bl
END;

Procedure Clear_Hires(Color: Byte); Assembler;
Asm
{ === Set color of pixel === }
    mov     dx,3ceh             { Graphic Controller address register port }
    mov     ax,0f01h            { Index 01h  (enable set/reset register) }
    out     dx,ax
    xor     al,al               { Index 00h  (set/reset register) }
    mov     ah,Color            { Color (0 to 15) }
    out     dx,ax

{ === Enable writebits === }
    mov     ah,11111111b        { All bits enabled }
    mov     al,08h              { Select Bit Mask Register }
    out     dx,ax

{ === Now clear the screen === }
    mov     ax,$A000
    mov     es,ax
    xor     di,di
    mov     cx,80*480
    mov     al,11111111b
    rep     stosb
End;

Procedure Set_Hires_Color(Color: Byte); Assembler;
Asm
    mov     dx,3ceh             { Graphic Controller address register port }
    mov     ax,0f01h            { Index 01h  (enable set/reset register) }
    out     dx,ax
    xor     al,al               { Index 00h  (set/reset register) }
    mov     ah,cl               { Color (0 to 15) }
    out     dx,ax               { Set color }
End;

Procedure Set_Lineair_Palette; Assembler;
Asm
    mov     ax,01013h           { Setup lineair palette (16 colour modes) }
    xor     bl,bl
    int     10h
End;

Procedure Move_To(X,Y: Word); Assembler;   { Used for hardware scrolling }
Asm
{ === Calculate vgapos === }
    mov     cx,Y                { cx=Y }
    mov     bx,cx               { bx=Y }
    shl     cx,6                { cx=cx*64 }
    shl     bx,4                { bx=bx*16 }
    add     cx,bx               { Y=Y*80  (80*8=640 pixels on 1 line) }
    mov     bx,X
    add     cx,bx               { cx=(Y*80)+X }

{ === Load into registers === }
    mov     dx,CrtC_Index
    mov     al,0ch              { Start Address High register }
    mov     ah,ch               { Send high byte of start address }
    out     dx,ax
    mov     al,0dh              { Start Address Low register }
    mov     ah,cl               { Send low byte of start address }
    out     dx,ax
End;

Procedure Set_Line(LineValue: Word); Assembler;  { Sets a splitline }
Asm
    mov     bx,LineValue        { Load value into bx }

{ === Set bits 0 to 7 === }
    mov     dx,CrtC_Index
    mov     ah,bl
    mov     al,18h              { Line Compare register }
    out     dx,ax               { Load first 8 bits (0..7) }

{ === Set 8th bit === }
    mov     al,07h              { Overflow register }
    out     dx,al
    inc     dx                  { Data register }
    in      al,dx               { Get byte }
    and     al,11101111b        { Disable 4th bit }
    mov     ah,bh               { Load high part into ah }
    and     ah,00000001b        { Test for bit 8 ! (is it enabled or not?) }
    shl     ah,4                { Shift to correct spot }
    or      al,ah               { 1 or 2 bits is/are 1? Resulting bit is 1 }
    out     dx,al               { Reprogram bit in register }

{ === Set 9th bit === }
    dec     dx                  { Set register }
    mov     al,09h              { Maximum Scan Line register }
    out     dx,al
    inc     dx
    in      al,dx               { Get byte }
    and     al,10111111b        { Disable 6th bit }
    mov     ah,bh               { Load high part into ah }
    and     ah,00000010b        { Check bit 9 ! (is it enabled or not?) }
    shl     ah,5                { Shift to correct spot }
    or      al,ah               { 1 or 2 bits is/are 1? Resulting bit is 1 }
    out     dx,al               { Tweak bit in register }
End;

Procedure Copy_Vga(Source,Dest,Count: Word); Assembler; { Use writemode 1! }
Asm
    push    ds
    mov     ax,0a000h           { Vga segment }
    mov     ds,ax
    mov     es,ax
    mov     si,Source           { ds:si => vga source }
    mov     di,Dest             { es:di => vga dest }
    mov     cx,Count
    rep     movsb               { Use bytes due to writemode 1 }
    pop     ds
End;

Procedure Set_Plane(Plane: Byte); Assembler;
Asm
    mov     dx,Seq_Index
    mov     al,02h              { Map Mask register }
    mov     ah,Plane            { Alter bits 0,1,2 and/or 3 in register }
    and     ah,00001111b
    out     dx,al               { Possible values: 0,1,2,4,8,15 }
End;

Procedure Set_Height(Height: Byte); Assembler; { Cellheight is 0,1,3,7 or 15 }
Asm
    mov     dx,CrtC_Index
    mov     al,09h
    mov     ah,Height
    and     ah,00001111b
    out     dx,ax
End;

Procedure Set_WriteMode(Mode: Byte); Assembler;
Asm
    mov     dx,Gfx_Index
    mov     al,05h              { Mode register }
    out     dx,al
    inc     dx
    in      al,dx               { Get byte }
    and     al,11111100b
    mov     ah,Mode             { Load mode }
    and     ah,00000011b
    or      al,ah               { Set correct bits }
    out     dx,al
End;

Procedure Set_Color(Color,R,G,B: Byte); Assembler;
Asm
    mov     dx,03c8h
    mov     al,Color            { Load colorvalue (0..255) to alter }
    out     dx,al
    inc     dx
    mov     al,R                { Red }
    out     dx,al
    mov     al,G                { Green }
    out     dx,al
    mov     al,B                { Blue }
    out     dx,al
End;

Procedure Get_Color(Color: Byte; Var R,G,B: Byte); Assembler; { NOT TESTED! }
Asm
    mov     dx,03c7h
    mov     al,Color
    out     dx,al               { Set color to alter }
    mov     dx,03c9h
    in      al,dx
    les     di,[R]              { Don't know if this works... }
    mov     [es:di],al
    in      al,dx
    les     di,[G]
    mov     [es:di],al
    in      al,dx
    les     di,[B]
    mov     [es:di],al
End;

Procedure Set_PaletteReg(Reg,Val: Byte); Assembler;
Asm
    mov     bl,Reg              { Register to alter }
    mov     bh,Val              { Value to load into register }
    mov     ax,1000h            { Function number }
    int     10h
End;

END.                            { End of GFX unit }



             { === Code by Vulture/Outlaw Triad 1996,1997,1998 === }