{
  OTG files (Outlaw Triad hires graphic file)

    Format of OTG file:
       - 54 byte header
       - compressed screendata (like pcx)

    OTG header:
       - ID                 (4 bytes)
       - Rgb values         (48 bytes)
       - Bytes per scanline (2 bytes)

  Decoding the compressed data is the same as a 16 color hires pcx file.
  The only different thing in an OTG file is the header.

    Load_Otg parameters:
       - Gfxfile: The file containing OTG data
       - Gfxpos: Where the data starts (for linking in .exe files)
       - Totalsize: # of bytes to decode
       - Vgapos: Where to display on vga
       - Fade: Whether or not to fade to correct rgb values

  Code by Vulture/Outlaw Triad (C) 1996
}

Unit Otg;                               { Unit to decode hires otg files }

INTERFACE

Const Buffer = 65528;                   { Largest possible }

Var   Temp: Pointer;                    { Memory pointer for data }
      HoriCount, Vga_Index, DataCount: Word;
      Mode,Bytes_Per_Line,Plane,Counter: Byte;
      Header: Array[0..53] of Byte;     { Array for header data }
      Palette: Array[0..47*2] of Byte;  { Array for palette data }

Procedure Decode_Otg;
Procedure Load_Otg(Gfxfile: String; GfxPos,TotalSize: LongInt; VgaPos: Word; Fade: Boolean);

IMPLEMENTATION

Procedure Set_Rgb_Palette; Assembler;   { Initialize color stuff }
Asm
{ === Init palette registers === }
   xor     dl,dl
   mov     cx,16
@PalRegLoop:
   mov     bl,dl                { Register }
   mov     bh,dl                { Value }
   mov     ax,1000h
   int     10h                  { Let bios handle this }
   inc     dl
   loop    @PalRegLoop          { Thus creating a lineair palette }

{ === Reset rgb values to 0 === }
   mov     dx,03c8h
   xor     al,al
   out     dx,al
   inc     dx
   mov     cx,48
@Resetloop:
   out     dx,al
   loop    @Resetloop
End;

Procedure Decode_Otg; Assembler;
Asm
   push    ds                   { Save stuff on stack }
   push    bp

{ === Init stuff for run through buffer === }
   mov     bl,Plane             { First plane }
   mov     bh,Bytes_Per_Line    { # of bytes on 1 scanline }
   mov     bp,DataCount         { # of bytes to process }
   xor     cx,cx
   mov     cl,Counter           { Single/multiple write }
   mov     dx,HoriCount         { Scanline pointer }
   mov     di,Vga_Index         { Where to start }
   cld                          { Clear direction flag }

{ === Set ds & es === }
   mov     ax,0a000h
   mov     es,ax                { Destination => es:di }
   lds     si,Temp              { Source => ds:si (ds=destoyed!) }

{ === Check on multiple write === }
   cmp     cl,0
   je      @Get_Data            { If cl = 0 then process next byte }
   lodsb                        { Else cl = counter so continue loop }
   jmp     @Write_Byte

{ === Start of decoding === }
@Get_Data:
   cmp     si,bp                { Check if more data is to be decoded }
   je      @Quit_Decode         { Quit on end of buffer or no more data }
   lodsb                        { Get byte from ds:si }

   cmp     al,11000000b         { Is it a loop byte? }
   jb      @Single_Write        { If not, write a single byte }
   mov     cl,al                { Else setup for multiple write }
   and     cl,00111111b         { Set loopcounter for multiple bytes }

   cmp     si,bp                { Range check }
   je      @Quit_Decode
   lodsb                        { Load next byte }

   jmp     @Write_Byte

@Single_Write:
   mov     cl,1                 { Set loopcounter for 1 byte }

@Write_Byte:                    { Write byte(s) to vga }
   stosb
   inc     dl                   { Increment collumn }
   cmp     dl,bh                { End of line? }
   je      @Reached_End         { If so, set other plane }
   loop    @Write_Byte          { If not, write another byte }
   jmp     @Get_Data            { If all bytes written, fetch new data }

@Reached_End:
   shl     bl,1                 { Next plane }
   cmp     bl,00001000b         { Is this the last plane? }
   jle     @Not_Last            { If not, reset scanline }
   shr     bl,4                 { If so, reset plane & next scanline }
   jmp     @Set_Plane

@Not_Last:
   sub     di,dx                { Reset to start of scanline }

@Set_Plane:
   push    ax                   { Save byte }
   cli
   mov     dx,03c4h
   mov     al,02h               { Map Mask Register }
   mov     ah,bl                { Load plane in ah }
   out     dx,ax
   sti
   pop     ax                   { Restore byte }
   xor     dx,dx                { Reset collumn counter }
   loop    @Write_Byte          { And write next byte or get data }
   jmp     @Get_Data

@Quit_Decode:

   pop     bp                   { Restore from stack }
   pop     ds

{ === Save variables === }
   mov     Plane,bl
   mov     Vga_Index,di
   mov     Counter,cl
   mov     HoriCount,dx
End;

Procedure Set_Otg_Palette; Assembler;  { Sets colors immidiately }
Asm
   push    ds

{ === Set ds:si === }
   lea     si,Header
   add     si,4                 { Pal data starts at offset 4 in header }

{ === Start at color 0 === }
   mov     dx,03c8h
   xor     al,al
   out     dx,al
   inc     dx

{ === Send 48 bytes === }
   mov     cx,48
@PalLoop:
   mov     al,byte ptr ds:[si]
   shr     al,2
   out     dx,al
   inc     si
   loop    @PalLoop

   pop     ds
End;

Procedure Fade_Otg_Palette; Assembler;   { Fades to correct colors }
Asm
   push    ds

{ === Set ds:si & es:di === }
   lea     si,Header
   add     si,4
   mov     ax,ds
   mov     es,ax
   lea     di,Palette

{ === Send 48 bytes === }
   mov     cx,48
@PalLoop:
   mov     al,byte ptr ds:[si]
   shr     al,2
   mov     byte ptr es:[di],al
   inc     di
   inc     si
   loop    @PalLoop

{ === Fade routine === }
   mov     cx,64
@Outerloop:
   push    cx

   lea     di,Palette
   lea     si,Palette+48
   mov     bx,48
@Innerloop:
   mov     al,byte ptr es:[di]
   mov     ah,byte ptr ds:[si]
   cmp     al,ah
   jl      @NoUpdate
   inc     ah
   mov     byte ptr ds:[si],ah
@NoUpdate:
   inc     di
   inc     si
   dec     bx
   jnz     @Innerloop

{ === Vertical retrace (three times) === }
   mov     cx,3
@Retrace:
   mov     dx,3dah
@Vrt:
   in      al,dx
   test    al,1000b
   jnz     @Vrt
@NoVrt:
   in      al,dx
   test    al,1000b
   jz      @NoVrt
   loop    @Retrace

{ === Set palette === }
   lea     si,Palette+48
   mov     dx,03c8h
   xor     al,al
   out     dx,al
   inc     dx
   mov     cx,48                        { 48 rgb values to update }
   rep     outsb
   pop     cx
   loop    @Outerloop                   { Loop 64 times }

   pop     ds
End;

Procedure Load_Otg(Gfxfile: String; GfxPos, TotalSize: LongInt; VgaPos: Word; Fade: Boolean);
Var F: File;                            { File }
    ReadCount: Word;
    SizeCounter: LongInt;
Begin

{ === Open file and set recordsize 1 === }
  {$I-}
  Assign(F, Gfxfile);
  Reset(F,1);
  Seek(F,GfxPos);                       { Seek start of data (for in .exe) }

{ === Read header into array === }
  BlockRead(F, Header, Sizeof(Header));

{ === Set various variables === }
  Set_Rgb_Palette;                      { Init palette stuff }
  Plane := 1;                           { First plane first... }
  Bytes_Per_Line := Header[52];         { Read # of bytes per scanline }
  Vga_Index := VgaPos;
  Counter := 0;
  SizeCounter := Sizeof(Header);

{ === Write to vga === }
  Getmem(Temp, Buffer);                 { Allocate memory }

  Repeat                                { Breaking the 64kB limit }
    If Totalsize < Buffer then ReadCount := Totalsize-Sizeof(Header)
                          else ReadCount := Buffer;
    BlockRead(F, Temp^, ReadCount, DataCount);
    Inc(SizeCounter,DataCount);
    Decode_Otg;
  Until SizeCounter = Totalsize;

{ === Free memory and close file === }
  Freemem(Temp, Buffer);                { Free allocated memory }

  If not Fade then Set_Otg_Palette      { Set colors right away }
  Else Fade_Otg_Palette;                { Fade to correct rgb values }

  Close(F);
  {$I+}
End;

END.